;----------------------;
; Exception handler(s) ;
;----------------------;

M       =       0
IRPC            N, 012345689ABCEF
EXC_0&N LABEL DWORD
        CLI
        PUSH            EBP
        MOV             EBP, ESP
        PUSHAD
        MOV             EAX, M
        M       =       M+1
        JMP             EXC_HANDLER
ENDM

M       =       010H
IRPC            N, 0123456789ABCDEF
EXC_1&N LABEL DWORD
        CLI
        PUSH            EBP
        MOV             EBP, ESP
        PUSHAD
        MOV             EAX, M
        M       =       M+1
        JMP             EXC_HANDLER
ENDM

EXC_07  LABEL DWORD
        CLTS
        IRETD

EXC_0D  PROC NEAR
        CLI

; LET'S CHECK THE "VM" FLAG AT THE "EFLAGS" REGISTER
        TEST            DWORD PTR SS:[ESP+3*4], 20000H  ; IS IT A V86 TASK?
        JZ              EXC_0D_PM                       ; NO, IT ISN'T

;--------------------------------------------------------------------------;
;                           V86 MONITOR STARTS HERE                        ;
;--------------------------------------------------------------------------;

; ESP CORRECTION BECAUSE OF THE ERROR CODE
        ADD             ESP, 4                          ; ESP = ^V86.EIP

; MAKING STACK FRAME AND STORING REGISTERS ON THE STACK
        PUSH            EBP
        MOV             EBP, ESP                        ; EBP+4 = ^V86.EIP
        @PUSH           <DS, ES, EAX, EBX, ECX, EDX>

        MOV             AX, FLAT_SEL
        MOV             ES, AX

; LET'S CHECK THE OPCODE AT THE ADDRESS WHERE INTERRUPT/EXCEPTION OCCURED

        MOVZX           EBX, WORD PTR SS:[EBP+2*4]      ; EBX = V86.CS
        CMP             BX, CODE_V86_IRET               ; EXIT FROM INT STUFF?
        JNE             EXC_0D_CONTINUE                 ; NO
;        JMP             RINT_EXIT                       ; YES
        JMP             IRQ_CONTINUE                    ; YES

EXC_0D_CONTINUE:
        SHL             EBX, 4                          ; EBX = V86.CS*16
        ADD             EBX, SS:[EBP+4]                 ; EBX = CS*16+IP

        MOV             AH, ES:[EBX]                    ; AH = V86's OPCODE

        MOV             AL, 3                           ; INT NUMBER = 3
        CMP             AH, 0CCH                        ; "INT 3"?
        JE              EXC_0D_PERFORM_V86_INT          ; YES

        MOV             AL, 4                           ; INT NUMBER = 4
        CMP             AH, 0CEH                        ; "INTO"?
        JE              EXC_0D_PERFORM_V86_INT          ; YES

        CMP             AH, 0CDH                        ; "INT ?"?
        JNE             EXC_0D_DO_NOT_PERFORM_V86_INT   ; NO
        MOV             AL, ES:[EBX+1]                  ; AL = INT NUMBER
        INC             WORD PTR SS:[EBP+4]             ; CORRECT V86.IP

EXC_0D_PERFORM_V86_INT:

        INC             WORD PTR SS:[EBP+4]             ; CORRECT V86.IP

        CMP             AL, 15H                         ; INT 15H?
        JNE             EXC_0D_NOT_15                   ; NO
        CMP             BYTE PTR SS:[ESP+3*4+1], 87H    ; FUNCTION 87H?
        JNE             EXC_0D_NOT_15

        INCLUDE         INT15F87.INC

        @POP            <EDX, ECX, EBX, EAX, ES, DS, EBP>

        IRETD

EXC_0D_NOT_15:

; PREPARING V86 STACK FOR RETURNING FROM THE INTERRUPT HANDLER

; LET'S MAKE SPACE FOR IP, CS AND FLAGS

        SUB             WORD PTR SS:[EBP+4*4], 6        ; V86.SP -= 6
        MOVZX           EBX, WORD PTR SS:[EBP+5*4]      ; EBX = V86.SS
        SHL             EBX, 4                          ; EBX = V86.SS*16
        ADD             EBX, SS:[EBP+4*4]               ; EBX = SS*16+SP

; LET'S PUT FLAGS TO THE V86 STACK AND ALTER FLAGS FOR INTERRUPT HANDLER

        MOV             ECX, SS:[EBP+3*4]               ; ECX = V86.EFLAGS
        MOV             ES:[EBX+4], CX                  ; FLAGS ARE PUT
        AND             CH, NOT 3                       ; CLEAR "IF" AND "TF"
        MOV             SS:[EBP+3*4], ECX               ; ALTERED FLAGS

; LET'S PUT CS AND IP TO THE V86's STACK (FOR IRET)

        MOV             CX, SS:[EBP+2*4]                ; CS FOR IRET
        MOV             ES:[EBX+2], CX
        MOV             CX, SS:[EBP+4]                  ; IP FOR IRET
        MOV             ES:[EBX], CX

; LET'S PUT INTERRUPT VECTOR ADDRESS

        MOVZX           EAX, AL                         ; AL = INT NUMBER
        SHL             EAX, 2
        MOV             EAX, ES:[EAX]                   ; EAX = RM INT VECTOR

        MOV             SS:[EBP+4], AX                  ; IP
        SHR             EAX, 16
        MOV             SS:[EBP+2*4], AX                ; CS

; GO AHEAD! RUN THE REAL MODE INTERRUPT HANDLER CODE!

        @POP            <EDX, ECX, EBX, EAX, ES, DS, EBP>
        IRETD

EXC_0D_DO_NOT_PERFORM_V86_INT:
        @POP            <EDX, ECX, EBX, EAX, ES, DS, EBP>
        SUB             ESP, 4
        MOV             DWORD PTR SS:[ESP], 0

EXC_0D_PM:
        PUSH            EBP
        MOV             EBP, ESP
        PUSHAD
        MOV             EAX, 0DH
        JMP             EXC_HANDLER
EXC_0D  ENDP

; It's a common part of all exception handlers
EXC_HANDLER     PROC    NEAR
        PUSH    DS

        MOV     BX, DS_SEL
        MOV     DS, BX

        CALL    CLRSCR
        LEA     ESI, MSG_EXC
        MOV     EDI, 40*2
        CALL    W_ASCIIZ                        ; "EXCEPTION: "
        ADD     EDI, 11*2
        CALL    W_EAX                           ; EXCEPTION NUMBER
        SUB     EDI, 11*2

        MOV     EBX, EBP
        ADD     EBX, 4                          ; EBX = ^CS:EIP

        MOVZX   EAX, BYTE PTR DS:[EXC_ERR+EAX]
        OR      EAX, EAX
        JZ      EXC_NO_ERR_CODE                 ; THE EXCEPTION HAS NO ERROR CODE

        ADD     EBX, 4                          ; CORRECT EBX

        LEA     ESI, MSG_EXC_ERR
        ADD     EDI, 80*2
        CALL    W_ASCIIZ                        ; "ERROR CODE: "

        MOV     EAX, SS:[EBP+4]
        AND     EAX, 0FFFFH
        ADD     EDI, 12*2
        CALL    W_EAX
        SUB     EDI, 12*2
        SUB     EDI, 80*2

EXC_NO_ERR_CODE:
        LEA     ESI, MSG_CSEIP
        ADD     EDI, 2*80*2
        CALL    W_ASCIIZ                        ; "CS=XXXXXXXX EIP=XXXXXXXX"
        ADD     EDI, 3*2
        MOVZX   EAX, WORD PTR SS:[EBX+4]
        CALL    W_EAX                           ; CS
        ADD     EDI, 13*2
        MOV     EAX, SS:[EBX]
        CALL    W_EAX                           ; EIP
        SUB     EDI, (3+13)*2

        ADD     EDI, 80*2
        LEA     ESI, MSG_CR0                    ; "CR0=XXXXXXXX"
        CALL    W_ASCIIZ
        ADD     EDI, 4*2
        MOV     EAX, CR0
        CALL    W_EAX                           ; CR0

;        jmp     $                               ; INFINITE LOOP

        MOV     AX, SS_SEL
        MOV     SS, AX
        MOV     ESP, DS:[ESP32]
        @RETF

        IRETD
EXC_HANDLER     ENDP
