;------;
; IRQs ;
;------;

M       =       0
IRPC            N, 0123456789ABCDEF                     ; 20...2F
IRQ_0&N LABEL DWORD
        PUSH            EBP
        MOV             EBP, ESP
        @PUSH           <DS, ES, FS, GS, SS>
        PUSHAD
        MOV             EAX, M
        M       =       M+1
        JMP             IRQ_HANDLER
ENDM

; It's a common part of all IRQ handlers
IRQ_HANDLER PROC NEAR
        XOR             EDX, EDX                ; IRQ HAS NO BEEN HANDLED YET
        MOV             BX, DS_SEL
        MOV             DS, BX
        MOV             EBX, DS:[IRQ_ADDR+EAX*4]
        OR              EBX, EBX
        JZ              IRQ_NO_PM_HANDLER       ; NO PROTECTED MODE HANDLER

        CALL            DWORD PTR DS:[IRQ_ADDR+EAX*4]
        OR              EDX, 1                  ; HANDLED IN PROTECTED MODE
IRQ_NO_PM_HANDLER:
        MOVZX           EBX, BYTE PTR DS:[PASS_IRQ+EAX]
        OR              EBX, EBX
        JZ              IRQ_NO_RM_HANDLER       ; IRQ IS NOT PASSED TO REAL MODE HANDLER

; DETERMINING DATA LENGTH IN STACK
        TEST            DWORD PTR SS:[EBP+3*4], 20000H  ; WAS IT A V86 TASK?
        JZ              IRQ_PM_TEST0            ; NO, IT WASN'T
        MOV             ECX, 9                  ; 9 EXTRA DWORDS ON STACK
        JMP             IRQ_LCONTINUE

IRQ_PM_TEST0:
        TEST            WORD PTR SS:[EBP+2*4], 3; WAS RPL > 0 ?
        JZ              IRQ_PM_TEST1            ; NO, IT WASN'T
        MOV             ECX, 5                  ; 5 EXTRA DWORDS ON STACK
        JMP             IRQ_LCONTINUE

IRQ_PM_TEST1:
        MOV             ECX, 3                  ; 3 EXTRA DWORDS ON STACK

IRQ_LCONTINUE:
        ADD             ECX, 1+5+8              ; PUSHED EBP,SEGS,32-BIT REGS
        LEA             EDI, NESTED_STACK
        ADD             EDI, DS:[NESTED_SP]
        ADD             DS:[NESTED_SP], 32*4    ; ALTERING "STACK POINTER"
        MOV             DS:[EDI], ECX           ; STORING NUMBER OF DWORDS
        ADD             EDI, 4

        MOV             ESI, ESP
IRQ_CYCLE:
        MOV             EBX, SS:[ESI]
        MOV             DS:[EDI], EBX
        ADD             ESI, 4
        ADD             EDI, 4
        LOOP            IRQ_CYCLE

        CMP             EAX, 8
        JNC             IRQ_INT70
        ADD             EAX, 8
        JMP             IRQ_INT8
IRQ_INT70:
        ADD             EAX, 68H
IRQ_INT8:
        MOV             EDI, EAX                ; EDI = INT NUMBER

; WORKING WITH V86 STACK

        MOV             AX, FLAT_SEL
        MOV             ES, AX

        XOR             EAX, EAX
        @PUSH           <EAX,EAX,EAX,EAX>

        XOR             EBX, EBX
        MOV             BX, STCK_INT_V86
        PUSH            EBX
        SHL             EBX, 4

        MOVZX           EAX, DS:[INT_SP]
        SUB             DS:[INT_SP], 1024
        SUB             EAX, 6
        PUSH            EAX

        ADD             EBX, EAX
        MOV             WORD PTR ES:[EBX], OFFSET V86_INT_X
        MOV             WORD PTR ES:[EBX+2], CODE_V86_IRET

        PUSHFD
        POP             EAX
        AND             AH, NOT 42H             ; CLEAR "NT" AND "IF"
        PUSH            EAX
        POPFD
        MOV             EAX, 23000H
        MOV             ES:[EBX+4], AX
        PUSH            EAX                             ; FLAGS

        MOV             EAX, EDI                        ; INT NUMBER

        SHL             EAX, 2
        MOV             EAX, ES:[EAX]
        MOV             EBX, EAX

        SHR             EAX, 16
        PUSH            EAX                             ; CS
        AND             EBX, 0FFFFH
        PUSH            EBX                             ; EIP

        IRETD

IRQ_CONTINUE:
        MOV             AX, DS_SEL
        MOV             DS, AX
        ADD             DS:[INT_SP], 1024
        SUB             DS:[NESTED_SP], 32*4

        LEA             ESI, NESTED_STACK
        ADD             ESI, DS:[NESTED_SP]
        MOV             ECX, DS:[ESI]

        ADD             ESI, 4
        MOV             SS, DS:[ESI+8*4]
        MOV             ESP, DS:[ESI+3*4]
        SUB             ESP, 8*4

        MOV             EDI, ESP
IRQ_CYCLE2:
        MOV             EBX, DS:[ESI]
        MOV             SS:[EDI], EBX
        ADD             ESI, 4
        ADD             EDI, 4
        LOOP            IRQ_CYCLE2

        OR              EDX, 2                  ; HANDLED IN REAL MODE
IRQ_NO_RM_HANDLER:
        TEST            EDX, 2
        JNZ             IRQ_EOI_SENT            ; EOI HAS BEEN SENT BY RM IRQ HANDLER

        MOV             AH, AL                  ; AH = IRQ NUMBER
        MOV             AL, EOI
        OUT             PORT_8259M, AL          ; MASTER PIC
        CMP             AH, 8                   ; IRQ0...IRQ7?
        JC              IRQ_EOI_SENT            ; YES
        OUT             PORT_8259S, AL          ; SLAVE PIC

IRQ_EOI_SENT:
        POPAD
        @POP            <SS, GS, FS, ES, DS, EBP>
        IRETD
IRQ_HANDLER ENDP
